from __future__ import annotations

import contextlib
import logging
import secrets
import subprocess
from collections.abc import Mapping
from typing import Generator

from ._juju import ConfigValue, Juju

logger = logging.getLogger('jubilant')


@contextlib.contextmanager
def temp_model(
    keep: bool = False,
    controller: str | None = None,
    cloud: str | None = None,
    config: Mapping[str, ConfigValue] | None = None,
    credential: str | None = None,
) -> Generator[Juju]:
    """Context manager to create a temporary model for running tests in.

    This creates a new model with a random name in the format ``jubilant-abcd1234``, and destroys
    it and its storage when the context manager exits.

    If destroying the model takes longer than 10 minutes, log an error and return successfully.
    This is mainly to work around issues with Microk8s destroy-model hanging indefinitely.

    Provides a :class:`Juju` instance to operate on.

    If you want to configure its parameters, such as ``wait_timeout``, set the appropriate
    attribute inside the ``with`` statement. For example, to create a pytest fixture using
    ``temp_model``::

        @pytest.fixture(scope='module')
        def juju():
            with jubilant.temp_model() as juju:
                juju.wait_timeout = 10 * 60
                yield juju  # run the test

    Args:
        keep: If true, keep the created model around when the context manager exits.
            For an example of a ``juju`` fixture that uses *keep*, see
            `Jubilant's conftest.py <https://github.com/canonical/jubilant/blob/main/tests/integration/conftest.py>`__.
        controller: Name of controller where the temporary model will be added.
        cloud: Name of cloud or region (or cloud/region) to use for the temporary model.
        config: Temporary model configuration as key-value pairs, for example,
            ``{'image-stream': 'daily'}``.
        credential: Name of cloud credential to use for the temporary model.
    """
    juju = Juju()
    model = 'jubilant-' + secrets.token_hex(4)  # 4 bytes (8 hex digits) should be plenty
    juju.add_model(model, cloud=cloud, controller=controller, config=config, credential=credential)
    try:
        yield juju
    finally:
        if not keep:
            assert juju.model is not None
            try:
                # We're not using juju.destroy_model() here, as Juju doesn't provide a way
                # to specify the timeout for the entire model destruction operation.
                args = ['destroy-model', juju.model, '--no-prompt', '--destroy-storage', '--force']
                juju._cli(*args, include_model=False, timeout=10 * 60)
                juju.model = None
            except subprocess.TimeoutExpired as exc:
                logger.error(
                    'timeout destroying model: %s\nStdout:\n%s\nStderr:\n%s',
                    exc,
                    exc.stdout,
                    exc.stderr,
                )



"""Dataclasses that contain parsed output from ``juju show-model --format=json``.

These dataclasses were originally `generated from <https://github.com/juju/juju/compare/main...benhoyt:juju:modelinfo-dataclasses-4>`_
the Go structs in the Juju codebase, to ensure they are correct. Class names
come from the Go struct name, whereas attribute names come from the JSON field
names.
"""

from __future__ import annotations

import dataclasses
from typing import Any


@dataclasses.dataclass(frozen=True)
class ModelCredential:
    name: str
    owner: str
    cloud: str

    validity_check: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> ModelCredential:
        return cls(
            name=d['name'],
            owner=d['owner'],
            cloud=d['cloud'],
            validity_check=d.get('validity-check') or '',
        )


@dataclasses.dataclass(frozen=True)
class ModelStatusInfo:
    current: str = ''
    message: str = ''
    reason: str = ''
    since: str = ''
    migration: str = ''
    migration_start: str = ''
    migration_end: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> ModelStatusInfo:
        return cls(
            current=d.get('current') or '',
            message=d.get('message') or '',
            reason=d.get('reason') or '',
            since=d.get('since') or '',
            migration=d.get('migration') or '',
            migration_start=d.get('migration-start') or '',
            migration_end=d.get('migration-end') or '',
        )


@dataclasses.dataclass(frozen=True)
class ModelUserInfo:
    access: str
    last_connection: str

    display_name: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> ModelUserInfo:
        return cls(
            access=d['access'],
            last_connection=d['last-connection'],
            display_name=d.get('display-name') or '',
        )


@dataclasses.dataclass(frozen=True)
class ModelMachineInfo:
    cores: int

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> ModelMachineInfo:
        return cls(
            cores=d['cores'],
        )


@dataclasses.dataclass(frozen=True)
class SecretBackendInfo:
    num_secrets: int
    status: str

    message: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> SecretBackendInfo:
        return cls(
            num_secrets=d['num-secrets'],
            status=d['status'],
            message=d.get('message') or '',
        )


@dataclasses.dataclass(frozen=True)
class SupportedFeature:
    name: str
    description: str

    version: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> SupportedFeature:
        return cls(
            name=d['name'],
            description=d['description'],
            version=d.get('version') or '',
        )


@dataclasses.dataclass(frozen=True)
class ModelInfo:
    """Parsed version of the object returned by ``juju show-model --format=json``."""

    name: str
    short_name: str
    model_uuid: str
    model_type: str
    controller_uuid: str
    controller_name: str
    is_controller: bool
    cloud: str
    life: str

    region: str = ''
    type: str = ''
    status: ModelStatusInfo = dataclasses.field(default_factory=ModelStatusInfo)
    users: dict[str, ModelUserInfo] = dataclasses.field(default_factory=dict)  # type: ignore
    machines: dict[str, ModelMachineInfo] = dataclasses.field(default_factory=dict)  # type: ignore
    secret_backends: dict[str, SecretBackendInfo] = dataclasses.field(default_factory=dict)  # type: ignore
    agent_version: str = ''
    credential: ModelCredential | None = None
    supported_features: list[SupportedFeature] = dataclasses.field(default_factory=list)  # type: ignore

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> ModelInfo:
        return cls(
            name=d['name'],
            short_name=d['short-name'],
            model_uuid=d['model-uuid'],
            model_type=d['model-type'],
            controller_uuid=d['controller-uuid'],
            controller_name=d['controller-name'],
            is_controller=d['is-controller'],
            cloud=d['cloud'],
            life=d['life'],
            region=d.get('region') or '',
            type=d.get('type') or '',
            status=ModelStatusInfo._from_dict(d['status']) if 'status' in d else ModelStatusInfo(),
            users={k: ModelUserInfo._from_dict(v) for k, v in d.get('users', {}).items()},
            machines={k: ModelMachineInfo._from_dict(v) for k, v in d.get('machines', {}).items()},
            secret_backends={
                k: SecretBackendInfo._from_dict(v) for k, v in d.get('secret-backends', {}).items()
            },
            agent_version=d.get('agent-version') or '',
            credential=ModelCredential._from_dict(d['credential']) if 'credential' in d else None,
            supported_features=[
                SupportedFeature._from_dict(x) for x in d.get('supported-features', [])
            ],
        )



from __future__ import annotations

import dataclasses
import re
from typing import Any

# Per https://github.com/juju/juju/blob/afed87bbe98189805bc81e1af5d5fac70195fc86/core/semversion/version.go#L86-L91,
# support version strings in the following formats ("1.2-release-arch" is not
# allowed in "strict" mode):
#
# 1.2.3-release-arch
# 1.2.3.4-release-arch
# 1.2-alpha3-release-arch
# 1.2-alpha3.4-release-arch
majmin = r'^(\d+)\.(\d+)'
relarch = r'([^-]+)-([^-]+)$'
majmin_patch_re = re.compile(rf'{majmin}\.(\d+)-{relarch}')
majmin_patch_build_re = re.compile(rf'{majmin}\.(\d+)\.(\d+)-{relarch}')
majmin_tag_re = re.compile(rf'{majmin}-([a-z]+)(\d+)-{relarch}')
majmin_tag_build_re = re.compile(rf'{majmin}-([a-z]+)(\d+)\.(\d+)-{relarch}')


# To avoid type checker issues with the Version.tuple property being named "tuple".
# See https://github.com/astral-sh/ty/issues/1747
_tuple = tuple


@dataclasses.dataclass(frozen=True)
class Version:
    """Parsed Juju CLI version as returned by ``juju version --format=json --all``.

    The version from Juju is normally in the form ``major.minor.patch-release-arch``.
    You can compare versions using the :attr:`tuple` property::

        if juju.version().tuple >= (3, 6, 11):
            ...  # Juju CLI is at least 3.6.11
    """

    major: int
    minor: int
    patch: int
    release: str = ''
    arch: str = ''

    tag: str | None = None
    build: int | None = None

    git_commit: str | None = None

    def __str__(self) -> str:
        """The string representation of the parsed version.

        This recreates the version string from Juju.
        """
        prefix = f'{self.major}.{self.minor}'
        if self.tag is not None:
            prefix += f'-{self.tag}{self.patch}'
        else:
            prefix += f'.{self.patch}'
        if self.build is not None:
            prefix += f'.{self.build}'
        return f'{prefix}-{self.release}-{self.arch}'

    @property
    def tuple(self) -> _tuple[int, int, int]:
        """The tuple ``(major, minor, patch)``."""
        return self.major, self.minor, self.patch

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> Version:
        version = d['version']
        tag = None
        build = None

        if match := majmin_patch_re.match(version):
            major, minor, patch, release, arch = match.groups()
        elif match := majmin_patch_build_re.match(version):
            major, minor, patch, build, release, arch = match.groups()
        elif match := majmin_tag_re.match(version):
            major, minor, tag, patch, release, arch = match.groups()
        elif match := majmin_tag_build_re.match(version):
            major, minor, tag, patch, build, release, arch = match.groups()
        else:
            raise ValueError(f'invalid version {version!r}')

        return Version(
            major=int(major),
            minor=int(minor),
            patch=int(patch),
            release=release,
            arch=arch,
            tag=tag,
            build=int(build) if build is not None else None,
            git_commit=d.get('git-commit'),
        )



"""Dataclasses that contain parsed output from ``juju status --format=json``.

These dataclasses were originally `generated from <https://github.com/juju/juju/compare/main...benhoyt:juju:status-dataclasses>`_
the Go structs in the Juju codebase, to ensure they are correct. Class names
come from the Go struct name, whereas attribute names come from the JSON field
names. The one exception is that "Application" has been renamed to "App"
throughout, for brevity (and "application" to "app").
"""

from __future__ import annotations

import dataclasses
from typing import Any

from . import _pretty

__all__ = [
    'AppStatus',
    'AppStatusRelation',
    'CombinedStorage',
    'ControllerStatus',
    'EntityStatus',
    'FilesystemAttachment',
    'FilesystemAttachments',
    'FilesystemInfo',
    'FormattedBase',
    'LxdProfileContents',
    'MachineStatus',
    'ModelStatus',
    'NetworkInterface',
    'OfferStatus',
    'RemoteAppStatus',
    'RemoteEndpoint',
    'Status',
    'StatusInfo',
    'StorageAttachments',
    'StorageInfo',
    'UnitStatus',
    'UnitStorageAttachment',
    'VolumeAttachment',
    'VolumeAttachments',
    'VolumeInfo',
]


@dataclasses.dataclass(frozen=True)
class FormattedBase:
    name: str
    channel: str

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> FormattedBase:
        return cls(
            name=d['name'],
            channel=d['channel'],
        )


@dataclasses.dataclass(frozen=True)
class StatusInfo:
    """The main status class used for application, unit, and machine status."""

    current: str = ''
    message: str = ''
    reason: str = ''
    since: str = ''
    version: str = ''
    life: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> StatusInfo:
        if 'status-error' in d:
            return cls(current='failed', message=d['status-error'])
        return cls(
            current=d.get('current') or '',
            message=d.get('message') or '',
            reason=d.get('reason') or '',
            since=d.get('since') or '',
            version=d.get('version') or '',
            life=d.get('life') or '',
        )


@dataclasses.dataclass(frozen=True)
class AppStatusRelation:
    related_app: str = ''
    interface: str = ''
    scope: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> AppStatusRelation:
        return cls(
            related_app=d.get('related-application') or '',
            interface=d.get('interface') or '',
            scope=d.get('scope') or '',
        )


@dataclasses.dataclass(frozen=True)
class UnitStatus:
    """Status of a single unit."""

    workload_status: StatusInfo = dataclasses.field(default_factory=StatusInfo)
    juju_status: StatusInfo = dataclasses.field(default_factory=StatusInfo)
    leader: bool = False
    upgrading_from: str = ''
    machine: str = ''
    open_ports: list[str] = dataclasses.field(default_factory=list)  # type: ignore
    public_address: str = ''
    address: str = ''
    provider_id: str = ''
    subordinates: dict[str, UnitStatus] = dataclasses.field(default_factory=dict)  # type: ignore

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> UnitStatus:
        if 'status-error' in d:
            return cls(
                workload_status=StatusInfo(current='failed', message=d['status-error']),
                juju_status=StatusInfo(current='failed', message=d['status-error']),
            )
        return cls(
            workload_status=(
                StatusInfo._from_dict(d['workload-status'])
                if 'workload-status' in d
                else StatusInfo()
            ),
            juju_status=(
                StatusInfo._from_dict(d['juju-status']) if 'juju-status' in d else StatusInfo()
            ),
            leader=d.get('leader') or False,
            upgrading_from=d.get('upgrading-from') or '',
            machine=d.get('machine') or '',
            open_ports=d.get('open-ports') or [],
            public_address=d.get('public-address') or '',
            address=d.get('address') or '',
            provider_id=d.get('provider-id') or '',
            subordinates={
                k: UnitStatus._from_dict(v) for k, v in d.get('subordinates', {}).items()
            },
        )

    @property
    def is_active(self) -> bool:
        """Report whether the workload status for this unit status is "active"."""
        return self.workload_status.current == 'active'

    @property
    def is_blocked(self) -> bool:
        """Report whether the workload status for this unit status is "blocked"."""
        return self.workload_status.current == 'blocked'

    @property
    def is_error(self) -> bool:
        """Report whether the workload status for this unit status is "error"."""
        return self.workload_status.current == 'error'

    @property
    def is_maintenance(self) -> bool:
        """Report whether the workload status for this unit status is "maintenance"."""
        return self.workload_status.current == 'maintenance'

    @property
    def is_waiting(self) -> bool:
        """Report whether the workload status for this unit status is "waiting"."""
        return self.workload_status.current == 'waiting'


@dataclasses.dataclass(frozen=True)
class AppStatus:
    """Status of a single application."""

    charm: str
    charm_origin: str
    charm_name: str
    charm_rev: int
    exposed: bool

    base: FormattedBase | None = None
    charm_channel: str = ''
    charm_version: str = ''
    charm_profile: str = ''
    can_upgrade_to: str = ''
    scale: int = 0
    provider_id: str = ''
    address: str = ''
    life: str = ''
    app_status: StatusInfo = dataclasses.field(default_factory=StatusInfo)
    relations: dict[str, list[AppStatusRelation]] = dataclasses.field(default_factory=dict)  # type: ignore
    subordinate_to: list[str] = dataclasses.field(default_factory=list)  # type: ignore
    units: dict[str, UnitStatus] = dataclasses.field(default_factory=dict)  # type: ignore
    version: str = ''
    endpoint_bindings: dict[str, str] = dataclasses.field(default_factory=dict)  # type: ignore

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> AppStatus:
        if 'status-error' in d:
            return cls(
                charm='<failed>',
                charm_origin='<failed>',
                charm_name='<failed>',
                charm_rev=-1,
                exposed=False,
                app_status=StatusInfo(current='failed', message=d['status-error']),
            )
        return cls(
            charm=d['charm'],
            charm_origin=d['charm-origin'],
            charm_name=d['charm-name'],
            charm_rev=d['charm-rev'],
            exposed=d['exposed'],
            base=FormattedBase._from_dict(d['base']) if 'base' in d else None,
            charm_channel=d.get('charm-channel') or '',
            charm_version=d.get('charm-version') or '',
            charm_profile=d.get('charm-profile') or '',
            can_upgrade_to=d.get('can-upgrade-to') or '',
            scale=d.get('scale') or 0,
            provider_id=d.get('provider-id') or '',
            address=d.get('address') or '',
            life=d.get('life') or '',
            app_status=(
                StatusInfo._from_dict(d['application-status'])
                if 'application-status' in d
                else StatusInfo()
            ),
            relations={
                k: [AppStatusRelation._from_dict(x) for x in v]
                for k, v in d.get('relations', {}).items()
            },
            subordinate_to=d.get('subordinate-to') or [],
            units={k: UnitStatus._from_dict(v) for k, v in d.get('units', {}).items()},
            version=d.get('version') or '',
            endpoint_bindings=d.get('endpoint-bindings') or {},
        )

    @property
    def is_active(self) -> bool:
        """Report whether the application status for this app is "active"."""
        return self.app_status.current == 'active'

    @property
    def is_blocked(self) -> bool:
        """Report whether the application status for this app is "blocked"."""
        return self.app_status.current == 'blocked'

    @property
    def is_error(self) -> bool:
        """Report whether the application status for this app is "error"."""
        return self.app_status.current == 'error'

    @property
    def is_maintenance(self) -> bool:
        """Report whether the application status for this app is "maintenance"."""
        return self.app_status.current == 'maintenance'

    @property
    def is_waiting(self) -> bool:
        """Report whether the application status for this app is "waiting"."""
        return self.app_status.current == 'waiting'


@dataclasses.dataclass(frozen=True)
class EntityStatus:
    """Status class used for storage status. See :class:`StatusInfo` for the main status class."""

    current: str = ''
    message: str = ''
    since: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> EntityStatus:
        return cls(
            current=d.get('current') or '',
            message=d.get('message') or '',
            since=d.get('since') or '',
        )


@dataclasses.dataclass(frozen=True)
class UnitStorageAttachment:
    machine: str = ''
    location: str = ''
    life: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> UnitStorageAttachment:
        return cls(
            machine=d.get('machine') or '',
            location=d.get('location') or '',
            life=d.get('life') or '',
        )


@dataclasses.dataclass(frozen=True)
class StorageAttachments:
    units: dict[str, UnitStorageAttachment]

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> StorageAttachments:
        return cls(
            units={k: UnitStorageAttachment._from_dict(v) for k, v in d['units'].items()},
        )


@dataclasses.dataclass(frozen=True)
class StorageInfo:
    kind: str
    status: EntityStatus
    persistent: bool

    life: str = ''
    attachments: StorageAttachments | None = None

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> StorageInfo:
        return cls(
            kind=d['kind'],
            status=EntityStatus._from_dict(d['status']),
            persistent=d['persistent'],
            life=d.get('life') or '',
            attachments=(
                StorageAttachments._from_dict(d['attachments']) if 'attachments' in d else None
            ),
        )


@dataclasses.dataclass(frozen=True)
class FilesystemAttachment:
    mount_point: str
    read_only: bool

    life: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> FilesystemAttachment:
        return cls(
            mount_point=d['mount-point'],
            read_only=d['read-only'],
            life=d.get('life') or '',
        )


@dataclasses.dataclass(frozen=True)
class FilesystemAttachments:
    machines: dict[str, FilesystemAttachment] = dataclasses.field(default_factory=dict)  # type: ignore
    containers: dict[str, FilesystemAttachment] = dataclasses.field(default_factory=dict)  # type: ignore
    units: dict[str, UnitStorageAttachment] = dataclasses.field(default_factory=dict)  # type: ignore

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> FilesystemAttachments:
        return cls(
            machines={
                k: FilesystemAttachment._from_dict(v) for k, v in d.get('machines', {}).items()
            },
            containers={
                k: FilesystemAttachment._from_dict(v) for k, v in d.get('containers', {}).items()
            },
            units={k: UnitStorageAttachment._from_dict(v) for k, v in d.get('units', {}).items()},
        )


@dataclasses.dataclass(frozen=True)
class FilesystemInfo:
    size: int

    provider_id: str = ''
    volume: str = ''
    storage: str = ''
    attachments: FilesystemAttachments = dataclasses.field(default_factory=FilesystemAttachments)
    pool: str = ''
    life: str = ''
    status: EntityStatus = dataclasses.field(default_factory=EntityStatus)

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> FilesystemInfo:
        return cls(
            size=d['size'],
            provider_id=d.get('provider-id') or '',
            volume=d.get('volume') or '',
            storage=d.get('storage') or '',
            attachments=(
                FilesystemAttachments._from_dict(d['attachments'])
                if 'attachments' in d
                else FilesystemAttachments()
            ),
            pool=d.get('pool') or '',
            life=d.get('life') or '',
            status=EntityStatus._from_dict(d['status']) if 'status' in d else EntityStatus(),
        )


@dataclasses.dataclass(frozen=True)
class VolumeAttachment:
    read_only: bool

    device: str = ''
    device_link: str = ''
    bus_address: str = ''
    life: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> VolumeAttachment:
        return cls(
            read_only=d['read-only'],
            device=d.get('device') or '',
            device_link=d.get('device-link') or '',
            bus_address=d.get('bus-address') or '',
            life=d.get('life') or '',
        )


@dataclasses.dataclass(frozen=True)
class VolumeAttachments:
    machines: dict[str, VolumeAttachment] = dataclasses.field(default_factory=dict)  # type: ignore
    containers: dict[str, VolumeAttachment] = dataclasses.field(default_factory=dict)  # type: ignore
    units: dict[str, UnitStorageAttachment] = dataclasses.field(default_factory=dict)  # type: ignore

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> VolumeAttachments:
        return cls(
            machines={k: VolumeAttachment._from_dict(v) for k, v in d.get('machines', {}).items()},
            containers={
                k: VolumeAttachment._from_dict(v) for k, v in d.get('containers', {}).items()
            },
            units={k: UnitStorageAttachment._from_dict(v) for k, v in d.get('units', {}).items()},
        )


@dataclasses.dataclass(frozen=True)
class VolumeInfo:
    size: int
    persistent: bool

    provider_id: str = ''
    storage: str = ''
    attachments: VolumeAttachments = dataclasses.field(default_factory=VolumeAttachments)
    pool: str = ''
    hardware_id: str = ''
    wwn: str = ''
    life: str = ''
    status: EntityStatus = dataclasses.field(default_factory=EntityStatus)

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> VolumeInfo:
        return cls(
            size=d['size'],
            persistent=d['persistent'],
            provider_id=d.get('provider-id') or '',
            storage=d.get('storage') or '',
            attachments=(
                VolumeAttachments._from_dict(d['attachments'])
                if 'attachments' in d
                else VolumeAttachments()
            ),
            pool=d.get('pool') or '',
            hardware_id=d.get('hardware-id') or '',
            wwn=d.get('wwn') or '',
            life=d.get('life') or '',
            status=EntityStatus._from_dict(d['status']) if 'status' in d else EntityStatus(),
        )


@dataclasses.dataclass(frozen=True)
class CombinedStorage:
    """Storage information."""

    storage: dict[str, StorageInfo] = dataclasses.field(default_factory=dict)  # type: ignore
    filesystems: dict[str, FilesystemInfo] = dataclasses.field(default_factory=dict)  # type: ignore
    volumes: dict[str, VolumeInfo] = dataclasses.field(default_factory=dict)  # type: ignore

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> CombinedStorage:
        return cls(
            storage={k: StorageInfo._from_dict(v) for k, v in d.get('storage', {}).items()},
            filesystems={
                k: FilesystemInfo._from_dict(v) for k, v in d.get('filesystems', {}).items()
            },
            volumes={k: VolumeInfo._from_dict(v) for k, v in d.get('volumes', {}).items()},
        )


@dataclasses.dataclass(frozen=True)
class ControllerStatus:
    """Basic controller information."""

    timestamp: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> ControllerStatus:
        return cls(
            timestamp=d.get('timestamp') or '',
        )


@dataclasses.dataclass(frozen=True)
class LxdProfileContents:
    config: dict[str, str]
    description: str
    devices: dict[str, dict[str, str]]

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> LxdProfileContents:
        return cls(
            config=d['config'],
            description=d['description'],
            devices=d['devices'],
        )


@dataclasses.dataclass(frozen=True)
class NetworkInterface:
    ip_addresses: list[str]
    mac_address: str
    is_up: bool

    gateway: str = ''
    dns_nameservers: list[str] = dataclasses.field(default_factory=list)  # type: ignore
    space: str = ''

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> NetworkInterface:
        return cls(
            ip_addresses=d['ip-addresses'],
            mac_address=d['mac-address'],
            is_up=d['is-up'],
            gateway=d.get('gateway') or '',
            dns_nameservers=d.get('dns-nameservers') or [],
            space=d.get('space') or '',
        )


@dataclasses.dataclass(frozen=True)
class MachineStatus:
    """Status of a single machine."""

    juju_status: StatusInfo = dataclasses.field(default_factory=StatusInfo)
    hostname: str = ''
    dns_name: str = ''
    ip_addresses: list[str] = dataclasses.field(default_factory=list)  # type: ignore
    instance_id: str = ''
    display_name: str = ''
    machine_status: StatusInfo = dataclasses.field(default_factory=StatusInfo)
    modification_status: StatusInfo = dataclasses.field(default_factory=StatusInfo)
    base: FormattedBase | None = None
    network_interfaces: dict[str, NetworkInterface] = dataclasses.field(default_factory=dict)  # type: ignore
    containers: dict[str, MachineStatus] = dataclasses.field(default_factory=dict)  # type: ignore
    constraints: str = ''
    hardware: str = ''
    controller_member_status: str = ''
    ha_primary: bool = False
    lxd_profiles: dict[str, LxdProfileContents] = dataclasses.field(default_factory=dict)  # type: ignore

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> MachineStatus:
        if 'status-error' in d:
            return cls(
                juju_status=StatusInfo(current='failed', message=d['status-error']),
                machine_status=StatusInfo(current='failed', message=d['status-error']),
            )
        return cls(
            juju_status=(
                StatusInfo._from_dict(d['juju-status']) if 'juju-status' in d else StatusInfo()
            ),
            hostname=d.get('hostname') or '',
            dns_name=d.get('dns-name') or '',
            ip_addresses=d.get('ip-addresses') or [],
            instance_id=d.get('instance-id') or '',
            display_name=d.get('display-name') or '',
            machine_status=(
                StatusInfo._from_dict(d['machine-status'])
                if 'machine-status' in d
                else StatusInfo()
            ),
            modification_status=(
                StatusInfo._from_dict(d['modification-status'])
                if 'modification-status' in d
                else StatusInfo()
            ),
            base=FormattedBase._from_dict(d['base']) if 'base' in d else None,
            network_interfaces={
                k: NetworkInterface._from_dict(v)
                for k, v in d.get('network-interfaces', {}).items()
            },
            containers={
                k: MachineStatus._from_dict(v) for k, v in d.get('containers', {}).items()
            },
            constraints=d.get('constraints') or '',
            hardware=d.get('hardware') or '',
            controller_member_status=d.get('controller-member-status') or '',
            ha_primary=d.get('ha-primary') or False,
            lxd_profiles={
                k: LxdProfileContents._from_dict(v) for k, v in d.get('lxd-profiles', {}).items()
            },
        )


@dataclasses.dataclass(frozen=True)
class ModelStatus:
    """Status and basic information about the model."""

    name: str
    """Name of model."""

    type: str
    """Type of model, for example, ``caas`` for a Kubernetes model."""

    controller: str
    """Name of controller."""

    cloud: str
    """Name of cloud, for example ``aws`` or ``microk8s``."""

    version: str
    """Juju agent version."""

    region: str = ''
    """Cloud region."""

    upgrade_available: str = ''
    """Version number if a new Juju agent is available."""

    model_status: StatusInfo = dataclasses.field(default_factory=StatusInfo)
    """Status of the model. Normally the *current* field is ``available``."""

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> ModelStatus:
        return cls(
            name=d['name'],
            type=d['type'],
            controller=d['controller'],
            cloud=d['cloud'],
            version=d['version'],
            region=d.get('region') or '',
            upgrade_available=d.get('upgrade-available') or '',
            model_status=(
                StatusInfo._from_dict(d['model-status']) if 'model-status' in d else StatusInfo()
            ),
        )


@dataclasses.dataclass(frozen=True)
class RemoteEndpoint:
    interface: str
    role: str

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> RemoteEndpoint:
        return cls(
            interface=d['interface'],
            role=d['role'],
        )


@dataclasses.dataclass(frozen=True)
class OfferStatus:
    app: str
    endpoints: dict[str, RemoteEndpoint]

    charm: str = ''
    total_connected_count: int = 0
    active_connected_count: int = 0

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> OfferStatus:
        if 'status-error' in d:
            return cls(app=f'<failed> ({d["status-error"]})', endpoints={})
        return cls(
            app=d['application'],
            endpoints={k: RemoteEndpoint._from_dict(v) for k, v in d['endpoints'].items()},
            charm=d.get('charm') or '',
            total_connected_count=d.get('total-connected-count') or 0,
            active_connected_count=d.get('active-connected-count') or 0,
        )


@dataclasses.dataclass(frozen=True)
class RemoteAppStatus:
    url: str

    endpoints: dict[str, RemoteEndpoint] = dataclasses.field(default_factory=dict)  # type: ignore
    life: str = ''
    app_status: StatusInfo = dataclasses.field(default_factory=StatusInfo)
    relations: dict[str, list[str]] = dataclasses.field(default_factory=dict)  # type: ignore

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> RemoteAppStatus:
        if 'status-error' in d:
            return cls(
                url='<failed>',
                app_status=StatusInfo(current='failed', message=d['status-error']),
            )
        return cls(
            url=d['url'],
            endpoints={k: RemoteEndpoint._from_dict(v) for k, v in d.get('endpoints', {}).items()},
            life=d.get('life') or '',
            app_status=(
                StatusInfo._from_dict(d['application-status'])
                if 'application-status' in d
                else StatusInfo()
            ),
            relations=d.get('relations') or {},
        )


@dataclasses.dataclass(frozen=True)
class Status:
    """Parsed version of the status object returned by ``juju status --format=json``."""

    model: ModelStatus
    """Model information."""

    machines: dict[str, MachineStatus]
    """Mapping of machine ID string (for example, ``"0"``) to machine information."""

    apps: dict[str, AppStatus]
    """Mapping of application name to application information."""

    app_endpoints: dict[str, RemoteAppStatus] = dataclasses.field(default_factory=dict)  # type: ignore
    """Mapping of offer name to remote application information."""

    offers: dict[str, OfferStatus] = dataclasses.field(default_factory=dict)  # type: ignore
    """Mapping of offer name to offer information."""

    storage: CombinedStorage = dataclasses.field(default_factory=CombinedStorage)
    """Storage information."""

    controller: ControllerStatus = dataclasses.field(default_factory=ControllerStatus)
    """Controller information."""

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> Status:
        return cls(
            model=ModelStatus._from_dict(d['model']),
            machines={k: MachineStatus._from_dict(v) for k, v in d['machines'].items()},
            apps={k: AppStatus._from_dict(v) for k, v in d['applications'].items()},
            app_endpoints={
                k: RemoteAppStatus._from_dict(v)
                for k, v in d.get('application-endpoints', {}).items()
            },
            offers={k: OfferStatus._from_dict(v) for k, v in d.get('offers', {}).items()},
            storage=(
                CombinedStorage._from_dict(d['storage']) if 'storage' in d else CombinedStorage()
            ),
            controller=(
                ControllerStatus._from_dict(d['controller'])
                if 'controller' in d
                else ControllerStatus()
            ),
        )

    def __repr__(self) -> str:
        """Return a pretty-printed version of the status."""
        return _pretty.dump(self)

    def __str__(self) -> str:
        """Return a pretty-printed version of the status."""
        return repr(self)

    def __eq__(self, other: object) -> bool:
        """Report whether two status objects are equivalent.

        This excludes the :attr:`controller` attribute, because that only has a timestamp that
        constantly updates.
        """
        if not isinstance(other, Status):
            return False
        for field in dataclasses.fields(self):
            if field.name == 'controller':
                continue
            if getattr(self, field.name) != getattr(other, field.name):
                return False
        return True

    def get_units(self, app: str) -> dict[str, UnitStatus]:
        """Get all units of the given *app*, including units of subordinate apps.

        For subordinate apps, this finds and returns the subordinate units using the app's
        ``subordinate_to`` list. For principal (non-subordinate) apps, this is equivalent to
        ``status.apps[app].units``.

        Returns:
            Dict of units where the key is the unit name and the value is the :class:`UnitStatus`.
            If *app* is not found, return an empty dict.
        """
        app_info = self.apps.get(app)
        if app_info is None:
            return {}
        if not app_info.subordinate_to:
            return app_info.units

        units: dict[str, UnitStatus] = {}
        app_prefix = app + '/'
        for principal in app_info.subordinate_to:
            for unit_info in self.apps[principal].units.values():
                for sub_name, sub in unit_info.subordinates.items():
                    if sub_name.startswith(app_prefix):
                        units[sub_name] = sub  # noqa: PERF403
        return units



"""Dataclasses that contain parsed output from juju secrets."""

from __future__ import annotations

import dataclasses
import datetime
from typing import Any


class SecretURI(str):
    """A string subclass that represents a secret URI ("secret:...")."""

    def __repr__(self):
        """Override to make the type obvious in error messages."""
        return f'SecretURI({super().__repr__()})'

    @property
    def unique_identifier(self) -> str:
        """Unique identifier of this secret URI.

        This is the secret's globally-unique identifier (currently a 20-character Xid,
        for example "9m4e2mr0ui3e8a215n4g").
        """
        if '/' in self:
            # Handle 'secret://MODEL-UUID/UNIQUE-IDENTIFIER'
            return self.rsplit('/', maxsplit=1)[-1]
        elif self.startswith('secret:'):
            # Handle common case of 'secret:UNIQUE-IDENTIFIER'
            return self[len('secret:') :]
        else:
            return str(self)


@dataclasses.dataclass(frozen=True)
class Secret:
    """Represents a secret."""

    uri: SecretURI
    revision: int
    owner: str
    created: datetime.datetime
    updated: datetime.datetime
    expires: str | None
    rotation: str | None
    rotates: datetime.datetime | None
    name: str | None
    label: str | None
    description: str | None
    access: list[Access] | None
    revisions: list[Revision] | None

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> Secret:
        return Secret(
            uri=SecretURI('secret:' + d['uri']),
            revision=d['revision'],
            owner=d['owner'],
            created=_datetime_from_iso(d['created']),
            updated=_datetime_from_iso(d['updated']),
            expires=d.get('expires'),
            rotation=d.get('rotation'),
            rotates=_datetime_from_iso(d['rotates']) if 'rotates' in d else None,
            name=d.get('name'),
            label=d.get('label'),
            description=d.get('description'),
            access=[Access._from_dict(access) for access in d.get('access', [])] or None,
            revisions=[Revision._from_dict(revision) for revision in d.get('revisions', [])]
            or None,
        )


@dataclasses.dataclass(frozen=True)
class RevealedSecret(Secret):
    """Represents a secret that was revealed, which has a content field that's populated."""

    checksum: str
    """Checksum of the secret value or an empty string for Juju controllers < 3.6.0."""

    content: dict[str, str]
    """Mapping of secret keys to secret values."""

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> RevealedSecret:
        kwargs = dataclasses.asdict(super()._from_dict(d))
        return RevealedSecret(
            # Secret content checksums were introduced in Juju 3.6.0
            content=d['content']['Data'],
            checksum=d.get('checksum', ''),
            **kwargs,
        )


@dataclasses.dataclass(frozen=True)
class Revision:
    """Represents a revision of a secret."""

    revision: int
    backend: str
    created: datetime.datetime
    updated: datetime.datetime
    expires: datetime.datetime | None

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> Revision:
        return cls(
            revision=d['revision'],
            backend=d.get('backend', ''),
            created=_datetime_from_iso(d['created']),
            updated=_datetime_from_iso(d['updated']),
            expires=_datetime_from_iso(d['expires']) if 'expires' in d else None,
        )


@dataclasses.dataclass(frozen=True)
class Access:
    """Represents access to a secret."""

    target: str
    scope: str
    role: str

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> Access:
        return cls(target=d['target'], scope=d['scope'], role=d['role'])


def _datetime_from_iso(dt: str) -> datetime.datetime:
    """Converts a Juju-specific ISO 8601 string to a datetime object."""
    return datetime.datetime.fromisoformat(dt.replace('Z', '+00:00'))
